/***************************************************************************
 *   Copyright (C) 2005 by Joel R. Voss                                    *
 *   jvoss@myuw.net                                                        *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include "neural.h"

// TODO: Neural Test Program wants extra verbosity.
#define NEURAL_TEST false
#ifndef NEURAL_TEST
#define NEURAL_TEST true
#endif // NEURAL_TEST

void NeuralNetwork_init(NeuralNetwork *net, WORD layers, WORD nodes)
{
	net->id   = 0;
	net->name = 0;
	net->num_layers = layers;
	if(NEURAL_TEST) printf("Neural Network: %x, %i layers, %i nodes, %i B.\n", net, layers, nodes, sizeof(NeuralNetwork));
	net->data = (NeuralLayer *)malloc(layers * sizeof(NeuralLayer));
	int i;
	for(i = 0; i < layers; i++)
	{
		NeuralLayer_init(&(net->data[i]), nodes);
	}
}

void NeuralNetwork_destroy(NeuralNetwork *net)
{
	net->id = 0;
	free(net->name);
	
	if(NEURAL_TEST) printf("Neural Network: %x, %i layers, %i B.\n", net, net->num_layers, sizeof(NeuralNetwork));
	int i;
	for(i = 0; i < net->num_layers; i++)
	{
		NeuralLayer_destroy(&(net->data[i]));
	}
	free(net->data);
	net->data = 0;
	net->num_layers = 0;
}

NeuralInput *NeuralNetwork_update(NeuralNetwork *net, NeuralInput *input)
{
	// Put each output data into the next layer's input.
	// Return the output layer's output.
	NeuralInput *inputN = input;
	NeuralInput *output = 0;
	int i;
	for(i = 0; i < net->num_layers; i++)
	{
		output = NeuralLayer_update(&(net->data[i]), inputN);
		// Be sure to free the intermediary data that NeuralLayer_update() mallocs.
		if(inputN != input) free(inputN);
		inputN = output;
	}
	if(NEURAL_TEST) printf("Updating network: %x, %i -> %s;\n", net, net->num_layers, (output[0] ? "true" : "false"));
	return output;
}

#include <time.h>
void NeuralNetwork_random(NeuralNetwork *net)
{
	// Randomize all the weights in such a way that the output is random.
	// This is useful to start testing various weights.
	time_t secs = time(0);
	srand(secs);
	int i;
	for(i = 0; i < net->num_layers; i++)
	{
		int j;
		NeuralLayer *layer = &(net->data[i]);
		for(j = 0; j < layer->num_nodes; j++)
		{
			int total = 0;
			int k;
			NeuralNode *node = &(layer->nodes[j]);
			for(k = 0; k < node->num_weights; k++)
			{
				node->weights[k] = rand();
				total += node->weights[k];
				if(NEURAL_TEST) printf("%i, ", node->weights[k]);
			}
			// It is usually good to have the bias around half total so that
			// firing half of the input neurons will make the neuron fire.
			node->bias = total / 2;
			if(NEURAL_TEST) printf(" ?> %i;\n", node->bias);
		}
	}
	if(NEURAL_TEST) printf("Randomized network: %x, %i;\n", net, secs);
}

void NeuralLayer_init(NeuralLayer *layer, WORD nodes)
{
	layer->num_nodes = nodes;
	if(NEURAL_TEST) printf("Neural Layer: %x, %i nodes, %i B.\n", layer, nodes, sizeof(NeuralLayer));
	layer->nodes = (NeuralNode *)malloc(nodes * sizeof(NeuralNode));
	int i;
	for(i = 0; i < nodes; i++)
	{
		NeuralNode_init(&(layer->nodes[i]), nodes);
	}
}

void NeuralLayer_destroy(NeuralLayer *layer)
{
	if(NEURAL_TEST) printf("Neural Layer: %x, %i nodes, %i B.\n", layer, layer->num_nodes, sizeof(NeuralLayer));
	int i;
	for(i = 0; i < layer->num_nodes; i++)
	{
		NeuralNode_destroy(&(layer->nodes[i]));
	}
	free(layer->nodes);
	layer->nodes = 0;
	layer->num_nodes = 0;
}

NeuralOutput *NeuralLayer_update(NeuralLayer *layer, NeuralInput *input)
{
	// Malloc an Array of Neural Inputs here.
	NeuralOutput *output = (NeuralInput *)malloc(layer->num_nodes * sizeof(NeuralInput));
	int i;
	for(i = 0; i < layer->num_nodes; i++)
	{
		output[i] = NeuralNode_update(&(layer->nodes[i]), input);
	}
	if(NEURAL_TEST) printf("Updating layer: %x, %i -> %s;\n", layer, layer->num_nodes, (output[0] ? "true" : "false"));
	// return it and force the caller to free it.
	return output;
}

void NeuralNode_init(NeuralNode *node, WORD weights)
{
	node->num_weights = weights;
	if(NEURAL_TEST) printf("Neural Node: %x, %i weights, %i B.\n", node, weights, sizeof(NeuralNode) + weights * sizeof(WORD));
	node->weights = (WORD *)malloc(weights * sizeof(WORD));
	int i;
	for(i = 0; i < weights; i++)
	{
		node->weights[i] = 0;
	}
}

void NeuralNode_destroy(NeuralNode *node)
{
	if(NEURAL_TEST) printf("Neural Node: %x, %i weights, %i B.\n", node, node->num_weights, sizeof(NeuralNode) + node->num_weights * sizeof(WORD));
	free(node->weights);
	node->weights = 0;
	node->num_weights = 0;
}

NeuralOutput NeuralNode_update(NeuralNode *node, NeuralInput *input)
{
	// Multiply each input with the weight and compare with the bias.
	int excitation = 0;
	int i;
	for(i = 0; i < node->num_weights; i++)
	{
		excitation += node->weights[i] * input[i];
	}
	NeuralOutput output = (excitation > node->bias);
	if(NEURAL_TEST) printf("Updating node: %x, %i -> %s;\n", node, node->num_weights, (output ? "true" : "false"));
	return output;
}

#ifdef _NEURAL_TEST


#endif // _NEURAL_TEST
